# Copyright 2019 Faculty Science Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import datetime

from pytz import UTC

from faculty.clients.experiment import (
    ExperimentRunStatus as FacultyExperimentRunStatus,
    LifecycleStage as FacultyLifecycleStage,
    Metric as FacultyMetric,
    Param as FacultyParam,
    Tag as FacultyTag,
)
from mlflow.entities import (
    Experiment,
    LifecycleStage,
    Metric,
    Param,
    Run,
    RunData,
    RunInfo,
    RunStatus,
    RunTag,
    ViewType,
)
from mlflow_faculty.py23 import to_timestamp
from mlflow.exceptions import MlflowException
from mlflow.utils.mlflow_tags import MLFLOW_RUN_NAME, MLFLOW_PARENT_RUN_ID

_FACULTY_TO_MLFLOW_RUN_STATUS_MAP = {
    FacultyExperimentRunStatus.RUNNING: "RUNNING",
    FacultyExperimentRunStatus.FINISHED: "FINISHED",
    FacultyExperimentRunStatus.FAILED: "FAILED",
    FacultyExperimentRunStatus.SCHEDULED: "SCHEDULED",
    FacultyExperimentRunStatus.KILLED: "KILLED",
}


# Below we are mapping both strings and RunStatus objects to Faculty RunStatus
# objects so we are less sensitive to future MLflow API changes
_MLFLOW_TO_FACULTY_RUN_STATUS_MAP = {
    "RUNNING": FacultyExperimentRunStatus.RUNNING,
    "FINISHED": FacultyExperimentRunStatus.FINISHED,
    "FAILED": FacultyExperimentRunStatus.FAILED,
    "SCHEDULED": FacultyExperimentRunStatus.SCHEDULED,
    "KILLED": FacultyExperimentRunStatus.KILLED,
    RunStatus.RUNNING: FacultyExperimentRunStatus.RUNNING,
    RunStatus.FINISHED: FacultyExperimentRunStatus.FINISHED,
    RunStatus.FAILED: FacultyExperimentRunStatus.FAILED,
    RunStatus.SCHEDULED: FacultyExperimentRunStatus.SCHEDULED,
    RunStatus.KILLED: FacultyExperimentRunStatus.KILLED,
}


_LIFECYCLE_STAGE_CONVERSION_MAP = {
    ViewType.ACTIVE_ONLY: FacultyLifecycleStage.ACTIVE,
    ViewType.DELETED_ONLY: FacultyLifecycleStage.DELETED,
    ViewType.ALL: None,
}


def _datetime_to_mlflow_timestamp(dt):
    return to_timestamp(dt) * 1000


def faculty_experiment_to_mlflow_experiment(faculty_experiment):
    active = faculty_experiment.deleted_at is None
    return Experiment(
        faculty_experiment.id,
        faculty_experiment.name,
        faculty_experiment.artifact_location,
        LifecycleStage.ACTIVE if active else LifecycleStage.DELETED,
    )


def faculty_http_error_to_mlflow_exception(faculty_http_error):
    return MlflowException(
        "{}. Received response {} with status code {}".format(
            faculty_http_error.error,
            faculty_http_error.error_code,
            faculty_http_error.response.status_code,
        )
    )


def faculty_run_to_mlflow_run(faculty_run):
    lifecycle_stage = (
        LifecycleStage.ACTIVE
        if faculty_run.deleted_at is None
        else LifecycleStage.DELETED
    )
    start_time = _datetime_to_mlflow_timestamp(faculty_run.started_at)
    end_time = (
        _datetime_to_mlflow_timestamp(faculty_run.ended_at)
        if faculty_run.ended_at is not None
        else None
    )

    tag_dict = {tag.key: tag.value for tag in faculty_run.tags}

    extra_mlflow_tags = []

    # Set run name tag if set as attribute but not already a tag
    if MLFLOW_RUN_NAME not in tag_dict and faculty_run.name:
        extra_mlflow_tags.append(RunTag(MLFLOW_RUN_NAME, faculty_run.name))

    # Set parent run ID tag if set as attribute but not already a tag
    if (
        MLFLOW_PARENT_RUN_ID not in tag_dict
        and faculty_run.parent_run_id is not None
    ):
        extra_mlflow_tags.append(
            RunTag(MLFLOW_PARENT_RUN_ID, faculty_run.parent_run_id.hex)
        )

    run_info = RunInfo(
        run_uuid=faculty_run.id.hex,
        experiment_id=faculty_run.experiment_id,
        user_id="",
        status=_FACULTY_TO_MLFLOW_RUN_STATUS_MAP[faculty_run.status],
        start_time=start_time,
        end_time=end_time,
        lifecycle_stage=lifecycle_stage,
        artifact_uri=faculty_run.artifact_location,
        run_id=faculty_run.id.hex,
    )
    run_data = RunData(
        params=[
            faculty_param_to_mlflow_param(param)
            for param in faculty_run.params
        ],
        metrics=[
            faculty_metric_to_mlflow_metric(metric)
            for metric in faculty_run.metrics
        ],
        tags=[faculty_tag_to_mlflow_tag(tag) for tag in faculty_run.tags]
        + extra_mlflow_tags,
    )
    run = Run(run_info, run_data)
    return run


def faculty_metric_to_mlflow_metric(faculty_metric):
    return Metric(
        key=faculty_metric.key,
        value=faculty_metric.value,
        timestamp=_datetime_to_mlflow_timestamp(faculty_metric.timestamp),
        step=faculty_metric.step,
    )


def mlflow_metric_to_faculty_metric(mlflow_metric):
    return FacultyMetric(
        key=mlflow_metric.key,
        value=mlflow_metric.value,
        timestamp=mlflow_timestamp_to_datetime(mlflow_metric.timestamp),
        step=mlflow_metric.step,
    )


def faculty_param_to_mlflow_param(faculty_param):
    return Param(key=faculty_param.key, value=faculty_param.value)


def mlflow_param_to_faculty_param(mlflow_param):
    return FacultyParam(key=mlflow_param.key, value=mlflow_param.value)


def mlflow_tag_to_faculty_tag(mlflow_tag):
    return FacultyTag(key=mlflow_tag.key, value=mlflow_tag.value)


def faculty_tag_to_mlflow_tag(faculty_tag):
    return RunTag(key=faculty_tag.key, value=faculty_tag.value)


def mlflow_timestamp_to_datetime(mlflow_timestamp):
    return datetime.fromtimestamp(mlflow_timestamp / 1000.0, tz=UTC)


def mlflow_viewtype_to_faculty_lifecycle_stage(mlflow_view_type):
    return _LIFECYCLE_STAGE_CONVERSION_MAP[mlflow_view_type]


def mlflow_to_faculty_run_status(mlflow_run_status):
    return _MLFLOW_TO_FACULTY_RUN_STATUS_MAP[mlflow_run_status]



from datetime import datetime

import six
from pytz import UTC

EPOCH = datetime(1970, 1, 1, tzinfo=UTC)


def to_timestamp(dt):
    if six.PY2:
        return (dt - EPOCH).total_seconds()
    else:
        return dt.timestamp()



# Copyright 2019 Faculty Science Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from uuid import UUID
from itertools import islice

from six.moves import urllib

import faculty
import faculty.clients.base
import faculty.clients.experiment
from faculty.clients.experiment import (
    ParamConflict as FacultyParamConflict,
    ExperimentDeleted as FacultyExperimentDeleted,
)
from mlflow.entities import ViewType
from mlflow.exceptions import MlflowException
from mlflow.store.abstract_store import AbstractStore
from mlflow.utils.mlflow_tags import MLFLOW_RUN_NAME, MLFLOW_PARENT_RUN_ID

from mlflow_faculty.converters import (
    faculty_experiment_to_mlflow_experiment,
    faculty_http_error_to_mlflow_exception,
    faculty_metric_to_mlflow_metric,
    faculty_run_to_mlflow_run,
    mlflow_timestamp_to_datetime,
    mlflow_metric_to_faculty_metric,
    mlflow_param_to_faculty_param,
    mlflow_tag_to_faculty_tag,
    mlflow_viewtype_to_faculty_lifecycle_stage,
    mlflow_to_faculty_run_status,
)


class FacultyRestStore(AbstractStore):
    def __init__(self, store_uri, **_):
        parsed_uri = urllib.parse.urlparse(store_uri)
        if parsed_uri.scheme != "faculty":
            raise ValueError("Not a faculty URI: {}".format(store_uri))
        # Test for PROJECT_ID in netloc rather than path.
        elif parsed_uri.netloc != "":
            raise ValueError(
                "Invalid URI {}. Netloc is reserved. "
                "Did you mean 'faculty:{}".format(store_uri, parsed_uri.netloc)
            )

        cleaned_path = parsed_uri.path.strip("/")
        try:
            self._project_id = UUID(cleaned_path)
        except ValueError:
            raise ValueError(
                "{} in given URI {} is not a valid UUID".format(
                    cleaned_path, store_uri
                )
            )

        self._client = faculty.client("experiment")

    def list_experiments(self, view_type=ViewType.ACTIVE_ONLY):
        """
        :param view_type: Qualify requested type of experiments.

        :return: a list of Experiment objects stored in store for requested
            view.
        """
        lifecycle_stage = mlflow_viewtype_to_faculty_lifecycle_stage(view_type)
        try:
            faculty_experiments = self._client.list(
                self._project_id, lifecycle_stage
            )
        except faculty.clients.base.HttpError as e:
            raise faculty_http_error_to_mlflow_exception(e)
        else:
            return [
                faculty_experiment_to_mlflow_experiment(faculty_experiment)
                for faculty_experiment in faculty_experiments
            ]

    def create_experiment(self, name, artifact_location):
        """
        Creates a new experiment.
        If an experiment with the given name already exists, throws exception.

        :param name: Desired name for an experiment
        :param artifact_location: Base location for artifacts in runs. May be
            None.

        :return: experiment_id (integer) for the newly created experiment.
        """
        try:
            faculty_experiment = self._client.create(
                self._project_id, name, artifact_location=artifact_location
            )
        except faculty.clients.experiment.ExperimentNameConflict as e:
            raise MlflowException(str(e))
        except faculty.clients.base.HttpError as e:
            raise faculty_http_error_to_mlflow_exception(e)
        else:
            return faculty_experiment.id

    def get_experiment(self, experiment_id):
        """
        Fetches the experiment by ID from the backend store.

        :param experiment_id: Integer id for the experiment

        :return: A single :py:class:`mlflow.entities.Experiment` object if it
            exists, otherwise raises an exception.
        """
        try:
            faculty_experiment = self._client.get(
                self._project_id, int(experiment_id)
            )
        except faculty.clients.base.HttpError as e:
            raise faculty_http_error_to_mlflow_exception(e)
        else:
            return faculty_experiment_to_mlflow_experiment(faculty_experiment)

    def get_experiment_by_name(self, experiment_name):
        """
        Fetches the experiment by name from the backend store.

        :param experiment_name: Name of experiment

        :return: A single :py:class:`mlflow.entities.Experiment` object if it
            exists.
        """
        return super(FacultyRestStore, self).get_experiment_by_name(
            experiment_name
        )

    def delete_experiment(self, experiment_id):
        """
        Deletes the experiment from the backend store. Deleted experiments can
        be restored until permanently deleted.

        :param experiment_id: Integer id for the experiment
        """
        try:
            self._client.delete(self._project_id, int(experiment_id))
        except faculty.clients.base.HttpError as e:
            raise faculty_http_error_to_mlflow_exception(e)

    def restore_experiment(self, experiment_id):
        """
        Restore deleted experiment unless it is permanently deleted.

        :param experiment_id: Integer id for the experiment
        """
        try:
            self._client.restore(self._project_id, int(experiment_id))
        except faculty.clients.base.HttpError as e:
            raise faculty_http_error_to_mlflow_exception(e)

    def rename_experiment(self, experiment_id, new_name):
        """
        Update an experiment's name. The new name must be unique.

        :param experiment_id: Integer id for the experiment
        """
        try:
            self._client.update(
                self._project_id, int(experiment_id), name=new_name
            )
        except faculty.clients.experiment.ExperimentNameConflict as e:
            raise MlflowException(str(e))
        except faculty.clients.base.HttpError as e:
            raise faculty_http_error_to_mlflow_exception(e)

    def get_run(self, run_id):
        """
        Fetches the run from backend store

        :param run_id: string containing run UUID
            (32 hex characters = a uuid4 stripped off of dashes)

        :return: A single :py:class:`mlflow.entities.Run` object if it exists,
            otherwise raises an exception
        """
        try:
            faculty_run = self._client.get_run(self._project_id, UUID(run_id))
        except faculty.clients.base.HttpError as e:
            raise faculty_http_error_to_mlflow_exception(e)
        else:
            mlflow_run = faculty_run_to_mlflow_run(faculty_run)
            return mlflow_run

    def update_run_info(self, run_id, run_status, end_time):
        """
        Updates the metadata of the specified run.

        :param run_id: string containing run UUID
        :param run_status: RunStatus to update the run to, optional
        :param end_time: timestamp to update the run ended_at to, optional

        :return: :py:class:`mlflow.entities.RunInfo` describing the updated
            run.
        """
        try:
            faculty_run = self._client.update_run_info(
                self._project_id,
                UUID(run_id),
                mlflow_to_faculty_run_status(run_status),
                mlflow_timestamp_to_datetime(end_time),
            )
        except faculty.clients.base.HttpError as e:
            raise faculty_http_error_to_mlflow_exception(e)
        else:
            mlflow_run = faculty_run_to_mlflow_run(faculty_run)
            return mlflow_run.info

    def create_run(self, experiment_id, user_id, start_time, tags):
        """
        Creates a run under the specified experiment ID, setting the run's
        status to "RUNNING" and the start time to the current time.

        :param experiment_id: ID of the experiment for this run
        :param user_id: ID of the user launching this run
        :param start_time: Time the run started at in epoch milliseconds.
        :param tags: List of Mlflow Tag entities.

        :return: The created Run object
        """
        tags = [] if tags is None else tags

        # For backward compatability, fall back to run name or parent run ID
        # set in tags
        tag_dict = {tag.key: tag.value for tag in tags}
        run_name = tag_dict.get(MLFLOW_RUN_NAME) or ""
        parent_run_id = tag_dict.get(MLFLOW_PARENT_RUN_ID) or None

        try:
            faculty_run = self._client.create_run(
                self._project_id,
                int(experiment_id),
                run_name,
                mlflow_timestamp_to_datetime(start_time),
                None if parent_run_id is None else UUID(parent_run_id),
                tags=[mlflow_tag_to_faculty_tag(tag) for tag in tags],
            )
        except FacultyExperimentDeleted as conflict:
            raise MlflowException(
                "Experiment {0} is deleted."
                " To create runs for this experiment,"
                " first restore it with the shell command "
                "'mlflow experiments restore {0}'".format(
                    conflict.experiment_id
                )
            )
        except faculty.clients.base.HttpError as e:
            raise faculty_http_error_to_mlflow_exception(e)
        else:
            mlflow_run = faculty_run_to_mlflow_run(faculty_run)
            return mlflow_run

    def delete_run(self, run_id):
        """
        Deletes a run.
        :param run_id:
        """

        run_id = UUID(run_id)

        try:
            response = self._client.delete_runs(self._project_id, [run_id])
        except faculty.clients.base.HttpError as e:
            raise faculty_http_error_to_mlflow_exception(e)

        if run_id in response.deleted_run_ids:
            return
        elif run_id in response.conflicted_run_ids:
            raise MlflowException(
                "Could not delete already-deleted run {}".format(run_id.hex)
            )
        else:
            raise MlflowException(
                "Could not delete non-existent run {}".format(run_id.hex)
            )

    def restore_run(self, run_id):
        """
        Restores a run.
        :param run_id:
        """

        run_id = UUID(run_id)

        try:
            response = self._client.restore_runs(self._project_id, [run_id])
        except faculty.clients.base.HttpError as e:
            raise faculty_http_error_to_mlflow_exception(e)

        if run_id in response.restored_run_ids:
            return
        elif run_id in response.conflicted_run_ids:
            raise MlflowException(
                "Could not restore already-active run {}".format(run_id.hex)
            )
        else:
            raise MlflowException(
                "Could not restore non-existent run {}".format(run_id.hex)
            )

    def get_metric_history(self, run_id, metric_key):
        """
        Returns all logged value for a given metric.

        :param run_id: Unique identifier for run
        :param metric_key: Metric name within the run

        :return: A list of float values logged for the give metric if logged,
            else empty list
        """
        try:
            metric_history = self._client.get_metric_history(
                self._project_id, UUID(run_id), metric_key
            )
        except faculty.clients.base.HttpError as e:
            raise faculty_http_error_to_mlflow_exception(e)
        else:
            return [
                faculty_metric_to_mlflow_metric(faculty_metric)
                for faculty_metric in metric_history
            ]

    def search_runs(
        self,
        experiment_ids,
        search_filter=None,
        run_view_type=None,
        max_results=None,
    ):
        """ Returns runs that match the given list of search expressions within
        the experiments.  Given multiple search expressions, all these
        expressions are ANDed together for search.

        :param experiment_ids: List[int] of experiment ids to scope the search
        :param search_filter: List of search expressions
        :param run_view_type: mlflow.entities.ViewType

        :return: A list of :py:class:`mlflow.entities.Run` objects that satisfy
            the search expressions
        """
        if search_filter is not None:
            raise NotImplementedError("search_expressions must be set to None")

        experiment_ids = (
            None if experiment_ids is None else list(map(int, experiment_ids))
        )

        def _get_runs():
            response = self._client.list_runs(
                self._project_id,
                experiment_ids=experiment_ids,
                lifecycle_stage=mlflow_viewtype_to_faculty_lifecycle_stage(
                    run_view_type
                ),
            )
            for run in response.runs:
                yield run
            next_page = response.pagination.next

            while next_page is not None:
                response = self._client.list_runs(
                    self._project_id,
                    experiment_ids=experiment_ids,
                    lifecycle_stage=mlflow_viewtype_to_faculty_lifecycle_stage(
                        run_view_type
                    ),
                    start=next_page.start,
                    limit=next_page.limit,
                )
                for run in response.runs:
                    yield run
                next_page = response.pagination.next

        try:
            run_generator = _get_runs()
            faculty_runs = list(islice(run_generator, max_results))
        except faculty.clients.base.HttpError as e:
            raise faculty_http_error_to_mlflow_exception(e)
        else:
            return [faculty_run_to_mlflow_run(run) for run in faculty_runs]

    def log_batch(self, run_id, metrics=None, params=None, tags=None):
        """
        Fetches the experiment by ID from the backend store.

        :param run_id: string containing run UUID
            (32 hex characters = a uuid4 stripped off of dashes)
        :param metrics: List of Mlflow Metric entities.
        :param params: List of Mlflow Param entities
        :param tags: List of Mlflow Tag entities.
        """
        metrics = [] if metrics is None else metrics
        params = [] if params is None else params
        tags = [] if tags is None else tags

        try:
            self._client.log_run_data(
                self._project_id,
                UUID(run_id),
                params=[
                    mlflow_param_to_faculty_param(param) for param in params
                ],
                metrics=[
                    mlflow_metric_to_faculty_metric(metric)
                    for metric in metrics
                ],
                tags=[mlflow_tag_to_faculty_tag(tag) for tag in tags],
            )
        except FacultyParamConflict as conflict:
            raise MlflowException(
                "Conflicting param keys: {}".format(
                    conflict.conflicting_params
                )
            )
        except faculty.clients.base.HttpError as e:
            raise faculty_http_error_to_mlflow_exception(e)



# Copyright 2019 Faculty Science Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import datetime
from uuid import uuid4

from faculty.clients.experiment import (
    Experiment as FacultyExperiment,
    ExperimentRun,
    ExperimentRunStatus,
    Metric as FacultyMetric,
    Param as FacultyParam,
    Tag as FacultyTag,
)
from mlflow.entities import (
    Experiment,
    LifecycleStage,
    Metric,
    Param,
    Run,
    RunTag,
    RunData,
    RunInfo,
)
from mlflow.utils.mlflow_tags import MLFLOW_RUN_NAME, MLFLOW_PARENT_RUN_ID
from pytz import UTC

from mlflow_faculty.py23 import to_timestamp


PROJECT_ID = uuid4()
EXPERIMENT_ID = 12

NAME = "experiment name"
ARTIFACT_LOCATION = "scheme://artifact-location"

METRIC_TIMESTAMP = datetime(2019, 3, 13, 17, 0, 15, tzinfo=UTC)
METRIC_TIMESTAMP_MILLISECONDS = to_timestamp(METRIC_TIMESTAMP) * 1000
FACULTY_METRIC = FacultyMetric(
    key="metric-key", value="metric-value", timestamp=METRIC_TIMESTAMP, step=0
)
MLFLOW_METRIC = Metric(
    "metric-key", "metric-value", METRIC_TIMESTAMP_MILLISECONDS, 0
)

FACULTY_PARAM = FacultyParam(key="param-key", value="param-value")
MLFLOW_PARAM = Param("param-key", "param-value")

FACULTY_TAG = FacultyTag(key="tag-key", value="tag-value")
MLFLOW_TAG = RunTag("tag-key", "tag-value")

FACULTY_EXPERIMENT = FacultyExperiment(
    id=EXPERIMENT_ID,
    name=NAME,
    description="not used",
    artifact_location=ARTIFACT_LOCATION,
    created_at=datetime.now(tz=UTC),
    last_updated_at=datetime.now(tz=UTC),
    deleted_at=None,
)

RUN_UUID = uuid4()
RUN_UUID_HEX_STR = RUN_UUID.hex

RUN_NUMBER = 42
RUN_NAME = "run name"

PARENT_RUN_UUID = uuid4()
PARENT_RUN_UUID_HEX_STR = PARENT_RUN_UUID.hex

RUN_STARTED_AT = datetime(2018, 3, 10, 11, 39, 12, 110000, tzinfo=UTC)
RUN_STARTED_AT_MILLISECONDS = to_timestamp(RUN_STARTED_AT) * 1000
RUN_ENDED_AT = datetime(2018, 3, 11, 11, 39, 12, 110000, tzinfo=UTC)
RUN_ENDED_AT_MILLISECONDS = to_timestamp(RUN_ENDED_AT) * 1000

FACULTY_RUN = ExperimentRun(
    id=RUN_UUID,
    run_number=RUN_NUMBER,
    experiment_id=FACULTY_EXPERIMENT.id,
    name=RUN_NAME,
    parent_run_id=PARENT_RUN_UUID,
    artifact_location=ARTIFACT_LOCATION,
    status=ExperimentRunStatus.RUNNING,
    started_at=RUN_STARTED_AT,
    ended_at=None,
    deleted_at=None,
    tags=[FACULTY_TAG],
    params=[FACULTY_PARAM],
    metrics=[FACULTY_METRIC],
)


def mlflow_experiment(lifecycle_stage=LifecycleStage.ACTIVE):
    return Experiment(EXPERIMENT_ID, NAME, ARTIFACT_LOCATION, lifecycle_stage)


def mlflow_run(
    name=RUN_NAME,
    status="RUNNING",
    end_time=None,
    lifecycle_stage=LifecycleStage.ACTIVE,
    name_tag=RunTag(MLFLOW_RUN_NAME, RUN_NAME),
    parent_run_id_tag=RunTag(MLFLOW_PARENT_RUN_ID, PARENT_RUN_UUID_HEX_STR),
):
    tags = [MLFLOW_TAG]
    if name_tag is not None:
        tags.append(name_tag)
    if parent_run_id_tag is not None:
        tags.append(parent_run_id_tag)
    data = RunData(params=[MLFLOW_PARAM], metrics=[MLFLOW_METRIC], tags=tags)
    info = RunInfo(
        run_uuid=RUN_UUID_HEX_STR,
        experiment_id=EXPERIMENT_ID,
        user_id="",
        status=status,
        start_time=RUN_STARTED_AT_MILLISECONDS,
        end_time=end_time,
        lifecycle_stage=lifecycle_stage,
        artifact_uri=ARTIFACT_LOCATION,
        run_id=RUN_UUID_HEX_STR,
    )
    return Run(info, data)



# Copyright 2019 Faculty Science Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import datetime

import pytest
from pytz import UTC
from requests import Response

from faculty.clients.experiment import (
    ExperimentRunStatus as FacultyExperimentRunStatus,
    LifecycleStage as FacultyLifecycleStage,
    Tag as FacultyTag,
)
from faculty.clients.base import HTTPError
from mlflow.exceptions import MlflowException
from mlflow.entities import LifecycleStage, RunTag, ViewType
from mlflow.utils.mlflow_tags import MLFLOW_RUN_NAME, MLFLOW_PARENT_RUN_ID

from mlflow_faculty.converters import (
    faculty_http_error_to_mlflow_exception,
    faculty_experiment_to_mlflow_experiment,
    faculty_metric_to_mlflow_metric,
    faculty_run_to_mlflow_run,
    mlflow_timestamp_to_datetime,
    faculty_tag_to_mlflow_tag,
    mlflow_metric_to_faculty_metric,
    mlflow_param_to_faculty_param,
    mlflow_tag_to_faculty_tag,
    mlflow_viewtype_to_faculty_lifecycle_stage,
)
from mlflow_faculty.py23 import to_timestamp
from tests.fixtures import (
    FACULTY_EXPERIMENT,
    FACULTY_METRIC,
    FACULTY_PARAM,
    FACULTY_RUN,
    FACULTY_TAG,
    MLFLOW_METRIC,
    MLFLOW_PARAM,
    MLFLOW_TAG,
    PARENT_RUN_UUID,
    PARENT_RUN_UUID_HEX_STR,
    mlflow_experiment,
    mlflow_run,
)

DATETIME = datetime(2018, 3, 10, 11, 45, 32, 110000, tzinfo=UTC)
DATETIME_MILLISECONDS = to_timestamp(DATETIME) * 1000


def experiment_equals(first, other):
    return (
        first.experiment_id == other.experiment_id
        and first.name == other.name
        and first.artifact_location == other.artifact_location
        and first.lifecycle_stage == other.lifecycle_stage
    )


def mlflow_object_equals(first, other):
    # mlflow objects return their properties as (key, value) when iterated over
    return all(x == y for x, y in zip(first, other))


def run_data_equals(first, other):
    return (
        first.metrics == other.metrics
        and first.params == other.params
        and first.tags == other.tags
    )


def run_equals(first, other):
    return run_data_equals(first.data, other.data) and first.info == other.info


def test_faculty_http_error_to_mlflow_exception():
    dummy_response = Response()
    dummy_response.status_code = 418
    faculty_http_error = HTTPError(dummy_response, "error", "error_code")

    assert isinstance(
        faculty_http_error_to_mlflow_exception(faculty_http_error),
        MlflowException,
    )


@pytest.mark.parametrize(
    "deleted_at, lifecycle_stage",
    [
        (None, LifecycleStage.ACTIVE),
        (datetime.now(tz=UTC), LifecycleStage.DELETED),
    ],
)
def test_faculty_experiment_to_mlflow_experiment(deleted_at, lifecycle_stage):
    faculty_experiment = FACULTY_EXPERIMENT._replace(deleted_at=deleted_at)
    expected_mlflow_experiment = mlflow_experiment(lifecycle_stage)

    assert experiment_equals(
        faculty_experiment_to_mlflow_experiment(faculty_experiment),
        expected_mlflow_experiment,
    )


@pytest.mark.parametrize(
    "faculty_run_status, mlflow_run_status",
    [
        (FacultyExperimentRunStatus.RUNNING, "RUNNING"),
        (FacultyExperimentRunStatus.FINISHED, "FINISHED"),
        (FacultyExperimentRunStatus.FAILED, "FAILED"),
        (FacultyExperimentRunStatus.SCHEDULED, "SCHEDULED"),
        (FacultyExperimentRunStatus.KILLED, "KILLED"),
    ],
    ids=["running", "finished", "failed", "scheduled", "killed"],
)
@pytest.mark.parametrize(
    "deleted_at, lifecycle_stage",
    [
        (None, LifecycleStage.ACTIVE),
        (datetime.now(tz=UTC), LifecycleStage.DELETED),
    ],
    ids=["active", "deleted"],
)
@pytest.mark.parametrize(
    "faculty_ended_at, mlflow_end_time",
    [(None, None), (DATETIME, DATETIME_MILLISECONDS)],
    ids=["not ended", "ended"],
)
def test_faculty_run_to_mlflow_run(
    faculty_run_status,
    mlflow_run_status,
    deleted_at,
    lifecycle_stage,
    faculty_ended_at,
    mlflow_end_time,
):
    faculty_run = FACULTY_RUN._replace(
        status=faculty_run_status,
        deleted_at=deleted_at,
        ended_at=faculty_ended_at,
    )

    expected_mlflow_run = mlflow_run(
        status=mlflow_run_status,
        end_time=mlflow_end_time,
        lifecycle_stage=lifecycle_stage,
    )

    assert run_equals(
        faculty_run_to_mlflow_run(faculty_run), expected_mlflow_run
    )


@pytest.mark.parametrize(
    "faculty_attribute, faculty_tag_value, mlflow_attribute, mlflow_tag_value",
    [
        ("attr", "tag", "attr", "tag"),
        ("attr", "", "attr", ""),
        ("attr", None, "attr", "attr"),
        ("", "tag", "tag", "tag"),
        ("", "", "", ""),
        ("", None, "", None),
    ],
)
def test_faculty_run_to_mlflow_run_name_backwards_compatability(
    faculty_attribute, faculty_tag_value, mlflow_attribute, mlflow_tag_value
):
    """Test logic setting run name tag when not available."""

    if faculty_tag_value is None:
        extra_tags = []
    else:
        extra_tags = [FacultyTag(MLFLOW_RUN_NAME, faculty_tag_value)]
    faculty_run = FACULTY_RUN._replace(
        name=faculty_attribute, tags=FACULTY_RUN.tags + extra_tags
    )

    if mlflow_tag_value is None:
        name_tag = None
    else:
        name_tag = RunTag(MLFLOW_RUN_NAME, mlflow_tag_value)
    expected_run = mlflow_run(name=mlflow_attribute, name_tag=name_tag)

    assert run_equals(faculty_run_to_mlflow_run(faculty_run), expected_run)


@pytest.mark.parametrize(
    "faculty_attribute, faculty_tag_value, mlflow_tag_value",
    [
        (PARENT_RUN_UUID, "other", "other"),
        (PARENT_RUN_UUID, "", ""),
        (PARENT_RUN_UUID, None, PARENT_RUN_UUID_HEX_STR),
        (None, PARENT_RUN_UUID_HEX_STR, PARENT_RUN_UUID_HEX_STR),
        (None, "", ""),
        (None, None, None),
    ],
)
def test_faculty_run_to_mlflow_run_parent_run_id_backwards_compatability(
    faculty_attribute, faculty_tag_value, mlflow_tag_value
):
    """Test logic setting parent run ID tag when not available."""

    if faculty_tag_value is None:
        extra_tags = []
    else:
        extra_tags = [FacultyTag(MLFLOW_PARENT_RUN_ID, faculty_tag_value)]
    faculty_run = FACULTY_RUN._replace(
        parent_run_id=faculty_attribute, tags=FACULTY_RUN.tags + extra_tags
    )

    if mlflow_tag_value is None:
        parent_run_id_tag = None
    else:
        parent_run_id_tag = RunTag(MLFLOW_PARENT_RUN_ID, mlflow_tag_value)
    expected_run = mlflow_run(parent_run_id_tag=parent_run_id_tag)

    assert run_equals(faculty_run_to_mlflow_run(faculty_run), expected_run)


def test_faculty_metric_to_mlflow_metric():
    assert mlflow_object_equals(
        faculty_metric_to_mlflow_metric(FACULTY_METRIC), MLFLOW_METRIC
    )


@pytest.mark.parametrize(
    "timestamp, expected_datetime",
    [
        (0, datetime(1970, 1, 1, tzinfo=UTC)),
        (1551884271987, datetime(2019, 3, 6, 14, 57, 51, 987000, tzinfo=UTC)),
    ],
)
def test_mlflow_timestamp_to_datetime(timestamp, expected_datetime):
    assert mlflow_timestamp_to_datetime(timestamp) == expected_datetime


def test_faculty_tag_to_mlflow_tag():
    assert faculty_tag_to_mlflow_tag(FACULTY_TAG) == MLFLOW_TAG


def test_mlflow_metric_to_faculty_metric():
    assert mlflow_metric_to_faculty_metric(MLFLOW_METRIC) == FACULTY_METRIC


def test_mlflow_param_to_faculty_param():
    assert mlflow_param_to_faculty_param(MLFLOW_PARAM) == FACULTY_PARAM


def test_mlflow_tag_to_faculty_tag():
    assert mlflow_tag_to_faculty_tag(MLFLOW_TAG) == FACULTY_TAG


def test_mlflow_viewtype_to_faculty_lifecycle_stage():
    assert mlflow_viewtype_to_faculty_lifecycle_stage(ViewType.ALL) is None
    assert (
        mlflow_viewtype_to_faculty_lifecycle_stage(ViewType.ACTIVE_ONLY)
        == FacultyLifecycleStage.ACTIVE
    )
    assert (
        mlflow_viewtype_to_faculty_lifecycle_stage(ViewType.DELETED_ONLY)
        == FacultyLifecycleStage.DELETED
    )



# Copyright 2019 Faculty Science Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import re

import faculty

# from mlflow.tracking.context import RunContextProvider


FACULTY_ENV_TAGS = [
    ("FACULTY_PROJECT_ID", "mlflow.faculty.project.projectId"),
    ("FACULTY_SERVER_ID", "mlflow.faculty.server.serverId"),
    ("FACULTY_SERVER_NAME", "mlflow.faculty.server.name"),
    ("NUM_CPUS", "mlflow.faculty.server.cpus"),
    ("AVAILABLE_MEMORY_MB", "mlflow.faculty.server.memoryMb"),
    ("NUM_GPUS", "mlflow.faculty.server.gpus"),
    ("FACULTY_JOB_ID", "mlflow.faculty.job.jobId"),
    ("FACULTY_JOB_NAME", "mlflow.faculty.job.name"),
    ("FACULTY_RUN_ID", "mlflow.faculty.job.runId"),
    ("FACULTY_RUN_NUMBER", "mlflow.faculty.job.runNumber"),
    ("FACULTY_SUBRUN_ID", "mlflow.faculty.job.subrunId"),
    ("FACULTY_SUBRUN_NUMBER", "mlflow.faculty.job.subrunNumber"),
    ("FACULTY_APP_ID", "mlflow.faculty.app.appId"),
    ("FACULTY_API_ID", "mlflow.faculty.api.apiId"),
]

USER_ID_TAG = "mlflow.faculty.user.userId"
USERNAME_TAG = "mlflow.faculty.user.username"
CREATED_BY_TAG = "mlflow.faculty.createdBy"
API_MODE_TAG = "mlflow.faculty.api.mode"


def _tags_from_account(account):
    return {USER_ID_TAG: str(account.user_id), USERNAME_TAG: account.username}


def _tags_from_server_type(server_type):
    if server_type is None:
        return {CREATED_BY_TAG: "user"}
    elif re.search("job", server_type):
        return {CREATED_BY_TAG: "job"}
    elif re.search("app", server_type):
        return {CREATED_BY_TAG: "app"}
    elif re.search("prod.*api", server_type):
        return {CREATED_BY_TAG: "api", API_MODE_TAG: "deploy"}
    elif re.search("dev.*api", server_type):
        return {CREATED_BY_TAG: "api", API_MODE_TAG: "test"}
    else:
        return {CREATED_BY_TAG: "user"}


class FacultyRunContext:  # TODO: This should inherit from RunContextProvider
    def __init__(self):
        self._account_cache = None

    def _get_account(self):
        if self._account_cache is None:
            client = faculty.client("account")
            self._account_cache = client.authenticated_account()
        return self._account_cache

    def in_context(self):
        return bool(os.environ.get("FACULTY_PROJECT_ID"))

    def tags(self):
        tags = {}

        for environment_variable, tag_name in FACULTY_ENV_TAGS:
            value = os.environ.get(environment_variable)
            if value:
                tags[tag_name] = value

        server_type = os.environ.get("FACULTY_SERVER_TYPE")
        tags.update(_tags_from_server_type(server_type))

        try:
            account = self._get_account()
        except Exception:
            pass
        else:
            tags.update(_tags_from_account(account))

        return tags



# Copyright 2019 Faculty Science Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from uuid import uuid4

import pytest

from mlflow_faculty.context import FacultyRunContext


def merge_dicts(*dicts):
    merged = {}
    for d in dicts:
        merged.update(d)
    return merged


USER_ID = uuid4()
USERNAME = "joe_bloggs"

STANDARD_ENVIRONMENT = {
    "FACULTY_PROJECT_ID": "project-id",
    "FACULTY_SERVER_ID": "server-id",
    "FACULTY_SERVER_NAME": "server-name",
    "FACULTY_SERVER_TYPE": "jupyter",
    "NUM_CPUS": "4",
    "AVAILABLE_MEMORY_MB": "4000",
    "NUM_GPUS": "1",
}
JOB_ENVIRONMENT = {
    "FACULTY_SERVER_TYPE": "python-job",
    "FACULTY_JOB_ID": "job-id",
    "FACULTY_JOB_NAME": "job-name",
    "FACULTY_RUN_ID": "run-id",
    "FACULTY_RUN_NUMBER": "23",
    "FACULTY_SUBRUN_ID": "subrun-id",
    "FACULTY_SUBRUN_NUMBER": "2",
}
APP_ENVIRONMENT = {"FACULTY_SERVER_TYPE": "app", "FACULTY_APP_ID": "app-id"}
DEPLOYED_API_ENVIRONMENT = {
    "FACULTY_SERVER_TYPE": "prod-python-api",
    "FACULTY_API_ID": "api-id",
}
TEST_API_ENVIRONMENT = {
    "FACULTY_SERVER_TYPE": "dev-python-api",
    "FACULTY_API_ID": "api-id",
}

ALL_ENVIRONMENT_EMPTY_STRING = {
    key: "" for key in merge_dicts(STANDARD_ENVIRONMENT, JOB_ENVIRONMENT)
}

DEFAULT_TAGS = {"mlflow.faculty.createdBy": "user"}
STANDARD_TAGS = {
    "mlflow.faculty.createdBy": "user",
    "mlflow.faculty.project.projectId": "project-id",
    "mlflow.faculty.server.serverId": "server-id",
    "mlflow.faculty.server.name": "server-name",
    "mlflow.faculty.server.cpus": "4",
    "mlflow.faculty.server.memoryMb": "4000",
    "mlflow.faculty.server.gpus": "1",
}
USER_TAGS = {
    "mlflow.faculty.user.userId": str(USER_ID),
    "mlflow.faculty.user.username": USERNAME,
}
JOB_TAGS = {
    "mlflow.faculty.createdBy": "job",
    "mlflow.faculty.job.jobId": "job-id",
    "mlflow.faculty.job.name": "job-name",
    "mlflow.faculty.job.runId": "run-id",
    "mlflow.faculty.job.runNumber": "23",
    "mlflow.faculty.job.subrunId": "subrun-id",
    "mlflow.faculty.job.subrunNumber": "2",
}
APP_TAGS = {
    "mlflow.faculty.createdBy": "app",
    "mlflow.faculty.app.appId": "app-id",
}
DEPLOYED_API_TAGS = {
    "mlflow.faculty.createdBy": "api",
    "mlflow.faculty.api.apiId": "api-id",
    "mlflow.faculty.api.mode": "deploy",
}
TEST_API_TAGS = {
    "mlflow.faculty.createdBy": "api",
    "mlflow.faculty.api.apiId": "api-id",
    "mlflow.faculty.api.mode": "test",
}


@pytest.mark.parametrize(
    "env, in_context",
    [
        ({"FACULTY_PROJECT_ID": "project-id"}, True),
        ({"FACULTY_PROJECT_ID": ""}, False),
        ({}, False),
    ],
)
def test_in_context(mocker, env, in_context):
    mocker.patch("os.environ", env)
    assert FacultyRunContext().in_context() is in_context


@pytest.mark.parametrize(
    "environment, environment_tags",
    [
        (STANDARD_ENVIRONMENT, STANDARD_TAGS),
        (
            merge_dicts(STANDARD_ENVIRONMENT, JOB_ENVIRONMENT),
            merge_dicts(STANDARD_TAGS, JOB_TAGS),
        ),
        (
            merge_dicts(STANDARD_ENVIRONMENT, APP_ENVIRONMENT),
            merge_dicts(STANDARD_TAGS, APP_TAGS),
        ),
        (
            merge_dicts(STANDARD_ENVIRONMENT, DEPLOYED_API_ENVIRONMENT),
            merge_dicts(STANDARD_TAGS, DEPLOYED_API_TAGS),
        ),
        (
            merge_dicts(STANDARD_ENVIRONMENT, TEST_API_ENVIRONMENT),
            merge_dicts(STANDARD_TAGS, TEST_API_TAGS),
        ),
        ({}, {}),
        (ALL_ENVIRONMENT_EMPTY_STRING, {}),
    ],
    ids=[
        "standard",
        "job",
        "app",
        "deployed-api",
        "test-api",
        "no-env",
        "env-empty-string",
    ],
)
@pytest.mark.parametrize(
    "account_available, user_tags",
    [(True, USER_TAGS), (False, {})],
    ids=["account", "no-account"],
)
def test_tags(
    mocker, environment, environment_tags, account_available, user_tags
):
    mocker.patch("os.environ", environment)

    mock_client = mocker.Mock()
    if account_available:
        account = mocker.Mock(user_id=USER_ID, username=USERNAME)
        mock_client.authenticated_account.return_value = account
    else:
        mock_client.authenticated_account.side_effect = Exception()
    mocker.patch("faculty.client", return_value=mock_client)

    expected_tags = merge_dicts(DEFAULT_TAGS, environment_tags, user_tags)

    assert FacultyRunContext().tags() == expected_tags



# Copyright 2019 Faculty Science Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from uuid import UUID
from functools import partial

from six.moves import urllib

from faculty.datasets import session
from mlflow.store.artifact_repo import ArtifactRepository
from mlflow.store.s3_artifact_repo import S3ArtifactRepository


class _S3ArtifactRepositoryWithClientOverride(S3ArtifactRepository):
    def __init__(self, artifact_uri, client_factory):
        super(_S3ArtifactRepositoryWithClientOverride, self).__init__(
            artifact_uri
        )
        self._client_factory = client_factory

    def _get_s3_client(self):
        return self._client_factory()


class FacultyDatasetsArtifactRepository(ArtifactRepository):
    def __init__(self, artifact_uri):
        parsed_uri = urllib.parse.urlparse(artifact_uri)
        if parsed_uri.scheme != "faculty-datasets":
            raise ValueError(
                "Not a Faculty datasets URI: {}".format(artifact_uri)
            )
        # Test for PROJECT_ID in netloc rather than path.
        elif parsed_uri.netloc != "":
            raise ValueError(
                "Invalid URI {}. Netloc is reserved. "
                "Did you mean 'faculty-datasets:{}{}'".format(
                    artifact_uri, parsed_uri.netloc, parsed_uri.path
                )
            )

        cleaned_path = parsed_uri.path.strip("/")
        first_part = cleaned_path.split("/")[0]
        try:
            project_id = UUID(first_part)
        except ValueError:
            raise ValueError(
                "{} in given URI {} is not a valid UUID".format(
                    first_part, artifact_uri
                )
            )

        self._session = session.get()
        s3_bucket = self._session.bucket(project_id)
        s3_uri = "s3://{}/{}".format(s3_bucket, cleaned_path)
        s3_client_factory = partial(self._session.s3_client, project_id)

        self._s3_repository = _S3ArtifactRepositoryWithClientOverride(
            s3_uri, s3_client_factory
        )

    def get_path_module(self):
        return self._s3_repository.get_path_module()

    def log_artifact(self, local_file, artifact_path=None):
        return self._s3_repository.log_artifact(local_file, artifact_path)

    def log_artifacts(self, local_dir, artifact_path=None):
        return self._s3_repository.log_artifacts(local_dir, artifact_path)

    def list_artifacts(self, path):
        return self._s3_repository.list_artifacts(path)

    def _download_file(self, remote_file_path, local_path):
        return self._s3_repository._download_file(remote_file_path, local_path)



from datetime import datetime

import pytest
from pytz import UTC

from mlflow_faculty.py23 import to_timestamp


@pytest.mark.parametrize(
    "dt, expected_timestamp",
    [
        (datetime(1970, 1, 1, tzinfo=UTC), 0),
        (datetime(2019, 3, 6, 14, 57, 51, tzinfo=UTC), 1551884271),
        (datetime(2019, 3, 6, 14, 57, 51, 987000, tzinfo=UTC), 1551884271.987),
    ],
)
def test_to_timestamp(dt, expected_timestamp):
    assert to_timestamp(dt) == expected_timestamp



# Copyright 2019 Faculty Science Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from setuptools import setup, find_packages


TRACKING_STORE_ENTRYPOINT = "faculty=mlflow_faculty:FacultyRestStore"
ARTIFACT_REPOSITORY_ENTRYPOINT = (
    "faculty-datasets=mlflow_faculty:FacultyDatasetsArtifactRepository"
)
RUN_CONTEXT_ENTRYPOINT = "faculty-run-context=mlflow_faculty:FacultyRunContext"


setup(
    name="mlflow-faculty",
    description="MLflow plugin for the Faculty platform.",
    url="https://faculty.ai/",
    author="Faculty",
    author_email="opensource@faculty.ai",
    license="Apache Software License",
    packages=find_packages(),
    use_scm_version={"version_scheme": "post-release"},
    setup_requires=["setuptools_scm"],
    install_requires=["mlflow>=1.0.0", "faculty>=0.24.0", "six", "pytz"],
    entry_points={
        "mlflow.tracking_store": TRACKING_STORE_ENTRYPOINT,
        "mlflow.artifact_repository": ARTIFACT_REPOSITORY_ENTRYPOINT,
        "mlflow.run_context_provider": RUN_CONTEXT_ENTRYPOINT,
    },
)



# Copyright 2019 Faculty Science Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License

import faculty
from faculty.clients.base import HttpError
from faculty.clients.experiment import (
    ExperimentNameConflict,
    ExperimentRunStatus as FacultyExperimentRunStatus,
    ListExperimentRunsResponse,
    DeleteExperimentRunsResponse,
    RestoreExperimentRunsResponse,
    Pagination,
    Page,
)
from mlflow.entities import RunStatus, RunTag, ViewType
from mlflow.exceptions import MlflowException
from mlflow.utils.mlflow_tags import MLFLOW_RUN_NAME, MLFLOW_PARENT_RUN_ID
import pytest

from mlflow_faculty.tracking import FacultyRestStore
from tests.fixtures import (
    ARTIFACT_LOCATION,
    EXPERIMENT_ID,
    RUN_ENDED_AT,
    RUN_ENDED_AT_MILLISECONDS,
    RUN_STARTED_AT_MILLISECONDS,
    RUN_UUID,
    RUN_UUID_HEX_STR,
    PARENT_RUN_UUID,
    PARENT_RUN_UUID_HEX_STR,
    FACULTY_EXPERIMENT,
    FACULTY_RUN,
    NAME,
    MLFLOW_METRIC,
    MLFLOW_PARAM,
    MLFLOW_TAG,
    PROJECT_ID,
)


STORE_URI = "faculty:{}".format(PROJECT_ID)


@pytest.mark.parametrize(
    "store_uri",
    [
        STORE_URI,
        "faculty:{}".format(PROJECT_ID),
        "faculty:/{}".format(PROJECT_ID),
        "faculty:///{}".format(PROJECT_ID),
        "faculty:///{}/".format(PROJECT_ID),
    ],
)
def test_init(mocker, store_uri):
    mocker.patch("faculty.client")

    store = FacultyRestStore(store_uri)

    assert store._project_id == PROJECT_ID
    faculty.client.assert_called_once_with("experiment")


def test_init_invalid_uri_scheme():
    store_uri = "invalid-scheme:/{}".format(PROJECT_ID)
    expected_error_message = "Not a faculty URI: {}".format(store_uri)
    with pytest.raises(ValueError, match=expected_error_message):
        FacultyRestStore(store_uri)


def test_init_invalid_uri_in_netloc():
    store_uri = "faculty://{}".format(PROJECT_ID)
    expected_error_message = (
        "Invalid URI {}. Netloc is reserved. "
        "Did you mean 'faculty:{}".format(store_uri, PROJECT_ID)
    )
    with pytest.raises(ValueError, match=expected_error_message):
        FacultyRestStore(store_uri)


def test_init_invalid_uri_bad_uuid():
    store_uri = "faculty:/invalid_uuid"
    expected_error = "invalid_uuid in given URI {} is not a valid UUID".format(
        store_uri
    )
    with pytest.raises(ValueError, match=expected_error):
        FacultyRestStore(store_uri)


def test_create_experiment(mocker):
    mock_client = mocker.Mock()
    mock_client.create.return_value = FACULTY_EXPERIMENT
    mocker.patch("faculty.client", return_value=mock_client)

    store = FacultyRestStore(STORE_URI)
    experiment_id = store.create_experiment(NAME, ARTIFACT_LOCATION)

    assert experiment_id == FACULTY_EXPERIMENT.id
    mock_client.create.assert_called_once_with(
        PROJECT_ID, NAME, artifact_location=ARTIFACT_LOCATION
    )


def test_create_experiment_name_conflict(mocker):
    exception = ExperimentNameConflict("bad name")
    mock_client = mocker.Mock()
    mock_client.create.side_effect = exception
    mocker.patch("faculty.client", return_value=mock_client)

    store = FacultyRestStore(STORE_URI)

    with pytest.raises(MlflowException, match=str(exception)):
        store.create_experiment(NAME, ARTIFACT_LOCATION)


def test_create_experiment_client_error(mocker):
    mock_client = mocker.Mock()
    mock_client.create.side_effect = HttpError(mocker.Mock(), "Error")
    mocker.patch("faculty.client", return_value=mock_client)

    store = FacultyRestStore(STORE_URI)

    with pytest.raises(MlflowException, match="Error"):
        store.create_experiment(NAME, ARTIFACT_LOCATION)


def test_get_experiment(mocker):
    faculty_experiment = mocker.Mock()
    mock_client = mocker.Mock()
    mock_client.get.return_value = faculty_experiment
    mocker.patch("faculty.client", return_value=mock_client)

    mlflow_experiment = mocker.Mock()
    converter = mocker.patch(
        "mlflow_faculty.tracking.faculty_experiment_to_mlflow_experiment",
        return_value=mlflow_experiment,
    )

    store = FacultyRestStore(STORE_URI)
    experiment = store.get_experiment(EXPERIMENT_ID)

    mock_client.get.assert_called_once_with(PROJECT_ID, EXPERIMENT_ID)
    converter.assert_called_once_with(faculty_experiment)
    assert experiment == mlflow_experiment


def test_get_experiment_client_error(mocker):
    mock_client = mocker.Mock()
    mock_client.get.side_effect = HttpError(
        mocker.Mock(), "Experiment with ID _ not found in project _"
    )
    mocker.patch("faculty.client", return_value=mock_client)

    store = FacultyRestStore(STORE_URI)

    with pytest.raises(
        MlflowException, match="Experiment with ID _ not found in project _"
    ):
        store.get_experiment(EXPERIMENT_ID)


def test_get_experiment_invalid_experiment_id(mocker):
    mock_client = mocker.Mock()
    mocker.patch("faculty.client", return_value=mock_client)

    store = FacultyRestStore(STORE_URI)

    with pytest.raises(ValueError):
        store.get_experiment("invalid-experiment-id")


def test_list_experiments(mocker):
    faculty_experiment = mocker.Mock()
    mock_client = mocker.Mock()
    mock_client.list.return_value = [faculty_experiment]
    mocker.patch("faculty.client", return_value=mock_client)

    mlflow_experiment = mocker.Mock()
    experiment_converter = mocker.patch(
        "mlflow_faculty.tracking.faculty_experiment_to_mlflow_experiment",
        return_value=mlflow_experiment,
    )
    lifecycle_stage = mocker.Mock()
    lifecycle_converter = mocker.patch(
        "mlflow_faculty.tracking.mlflow_viewtype_to_faculty_lifecycle_stage",
        return_value=lifecycle_stage,
    )

    store = FacultyRestStore(STORE_URI)
    experiments = store.list_experiments()

    mock_client.list.assert_called_once_with(PROJECT_ID, lifecycle_stage)
    experiment_converter.assert_called_once_with(faculty_experiment)
    lifecycle_converter.assert_called_once_with(ViewType.ACTIVE_ONLY)
    assert experiments == [mlflow_experiment]


def test_list_experiments_filtered_by_lifecycle_stage(mocker):
    faculty_experiment = mocker.Mock()
    mock_client = mocker.Mock()
    mock_client.list.return_value = [faculty_experiment]
    mocker.patch("faculty.client", return_value=mock_client)

    mlflow_experiment = mocker.Mock()
    experiment_converter = mocker.patch(
        "mlflow_faculty.tracking.faculty_experiment_to_mlflow_experiment",
        return_value=mlflow_experiment,
    )
    lifecycle_stage = mocker.Mock()
    lifecycle_converter = mocker.patch(
        "mlflow_faculty.tracking.mlflow_viewtype_to_faculty_lifecycle_stage",
        return_value=lifecycle_stage,
    )

    store = FacultyRestStore(STORE_URI)
    experiments = store.list_experiments(ViewType.DELETED_ONLY)

    mock_client.list.assert_called_once_with(PROJECT_ID, lifecycle_stage)
    experiment_converter.assert_called_once_with(faculty_experiment)
    lifecycle_converter.assert_called_once_with(ViewType.DELETED_ONLY)
    assert experiments == [mlflow_experiment]


def test_list_experiments_client_error(mocker):
    mock_client = mocker.Mock()
    mock_client.list.side_effect = HttpError(mocker.Mock(), "Error")
    mocker.patch("faculty.client", return_value=mock_client)

    store = FacultyRestStore(STORE_URI)

    with pytest.raises(MlflowException, match="Error"):
        store.list_experiments()


def test_rename_experiment(mocker):
    mock_client = mocker.Mock()
    mocker.patch("faculty.client", return_value=mock_client)

    store = FacultyRestStore(STORE_URI)
    store.rename_experiment(EXPERIMENT_ID, "new name")

    mock_client.update.assert_called_once_with(
        PROJECT_ID, EXPERIMENT_ID, name="new name"
    )


def test_rename_experiment_name_conflict(mocker):
    exception = ExperimentNameConflict("bad name")
    mock_client = mocker.Mock()
    mock_client.update.side_effect = exception
    mocker.patch("faculty.client", return_value=mock_client)

    store = FacultyRestStore(STORE_URI)

    with pytest.raises(MlflowException, match=str(exception)):
        store.rename_experiment(EXPERIMENT_ID, "bad name")


def test_rename_experiment_client_error(mocker):
    mock_client = mocker.Mock()
    mock_client.update.side_effect = HttpError(mocker.Mock(), "Error")
    mocker.patch("faculty.client", return_value=mock_client)

    store = FacultyRestStore(STORE_URI)

    with pytest.raises(MlflowException, match="Error"):
        store.rename_experiment(EXPERIMENT_ID, "new name")


def test_rename_experiment_invalid_experiment_id(mocker):
    mock_client = mocker.Mock()
    mocker.patch("faculty.client", return_value=mock_client)

    store = FacultyRestStore(STORE_URI)

    with pytest.raises(ValueError):
        store.rename_experiment("invalid-experiment-id", "new name")


def test_create_run(mocker):
    mlflow_timestamp = mocker.Mock()
    faculty_datetime = mocker.Mock()
    timestamp_converter_mock = mocker.patch(
        "mlflow_faculty.tracking.mlflow_timestamp_to_datetime",
        return_value=faculty_datetime,
    )

    mlflow_tag = mocker.Mock()
    faculty_tag = mocker.Mock()
    tag_converter_mock = mocker.patch(
        "mlflow_faculty.tracking.mlflow_tag_to_faculty_tag",
        return_value=faculty_tag,
    )

    faculty_run = mocker.Mock()
    mock_client = mocker.Mock()
    mock_client.create_run.return_value = faculty_run
    mocker.patch("faculty.client", return_value=mock_client)

    mlflow_run = mocker.Mock()
    run_converter_mock = mocker.patch(
        "mlflow_faculty.tracking.faculty_run_to_mlflow_run",
        return_value=mlflow_run,
    )

    store = FacultyRestStore(STORE_URI)

    returned_run = store.create_run(
        EXPERIMENT_ID, "unused-mlflow-user-id", mlflow_timestamp, [mlflow_tag]
    )

    timestamp_converter_mock.assert_called_once_with(mlflow_timestamp)
    tag_converter_mock.assert_called_once_with(mlflow_tag)
    mock_client.create_run.assert_called_once_with(
        PROJECT_ID,
        EXPERIMENT_ID,
        "",
        faculty_datetime,
        None,
        tags=[faculty_tag],
    )
    run_converter_mock.assert_called_once_with(faculty_run)
    assert returned_run == mlflow_run


def test_create_run_experiment_deleted(mocker):
    mocker.patch(
        "mlflow_faculty.tracking.mlflow_timestamp_to_datetime",
        return_value=mocker.Mock(),
    )

    mocker.patch(
        "mlflow_faculty.tracking.mlflow_tag_to_faculty_tag",
        return_value=mocker.Mock(),
    )

    mock_client = mocker.Mock()
    exception = faculty.clients.experiment.ExperimentDeleted(
        message="message", experiment_id="test-id"
    )
    mock_client.create_run.side_effect = exception

    mocker.patch("faculty.client", return_value=mock_client)

    store = FacultyRestStore(STORE_URI)

    with pytest.raises(MlflowException, match="experiment"):
        store.create_run(
            EXPERIMENT_ID,
            "unused-mlflow-user-id",
            RUN_STARTED_AT_MILLISECONDS,
            tags=[],
        )


@pytest.mark.parametrize(
    "run_name_tag, expected_run_name",
    [("tag name", "tag name"), ("", ""), (None, "")],
)
@pytest.mark.parametrize(
    "parent_run_id_tag, expected_parent_run_id",
    [(PARENT_RUN_UUID_HEX_STR, PARENT_RUN_UUID), ("", None), (None, None)],
)
def test_create_run_backwards_compatability(
    mocker,
    run_name_tag,
    expected_run_name,
    parent_run_id_tag,
    expected_parent_run_id,
):
    mocker.patch("mlflow_faculty.tracking.mlflow_timestamp_to_datetime")
    mocker.patch("mlflow_faculty.tracking.mlflow_tag_to_faculty_tag")

    mock_client = mocker.Mock()
    mocker.patch("faculty.client", return_value=mock_client)

    mocker.patch("mlflow_faculty.tracking.faculty_run_to_mlflow_run")

    tags = []
    if run_name_tag is not None:
        tags.append(RunTag(key=MLFLOW_RUN_NAME, value=run_name_tag))
    if parent_run_id_tag is not None:
        tags.append(RunTag(key=MLFLOW_PARENT_RUN_ID, value=parent_run_id_tag))

    store = FacultyRestStore(STORE_URI)

    store.create_run(
        EXPERIMENT_ID,
        "unused-mlflow-user-id",
        RUN_STARTED_AT_MILLISECONDS,
        tags,
    )

    args, _ = mock_client.create_run.call_args
    assert args[2] == expected_run_name
    assert args[4] == expected_parent_run_id


def test_create_run_client_error(mocker):
    mocker.patch("mlflow_faculty.tracking.mlflow_timestamp_to_datetime")
    mocker.patch("mlflow_faculty.tracking.mlflow_tag_to_faculty_tag")

    mock_client = mocker.Mock()
    mock_client.create_run.side_effect = HttpError(mocker.Mock(), "Some error")
    mocker.patch("faculty.client", return_value=mock_client)

    store = FacultyRestStore(STORE_URI)

    with pytest.raises(MlflowException, match="Some error"):
        store.create_run(
            EXPERIMENT_ID,
            "unused-mlflow-user-id",
            RUN_STARTED_AT_MILLISECONDS,
            tags=[],
        )


def test_create_run_invalid_experiment_id(mocker):
    mock_client = mocker.Mock()
    mocker.patch("faculty.client", return_value=mock_client)

    store = FacultyRestStore(STORE_URI)

    with pytest.raises(ValueError):
        store.create_run(
            "invalid-experiment-id",
            "unused-mlflow-user-id",
            RUN_STARTED_AT_MILLISECONDS,
            tags=[],
        )


def test_create_run_invalid_parent_run_id(mocker):
    mock_client = mocker.Mock()
    mocker.patch("faculty.client", return_value=mock_client)

    store = FacultyRestStore(STORE_URI)

    with pytest.raises(ValueError):
        store.create_run(
            EXPERIMENT_ID,
            "unused-mlflow-user-id",
            RUN_STARTED_AT_MILLISECONDS,
            [RunTag(key=MLFLOW_PARENT_RUN_ID, value="invalid-uuid")],
        )


def test_get_run(mocker):
    mock_client = mocker.Mock()
    mock_client.get_run.return_value = FACULTY_RUN
    mocker.patch("faculty.client", return_value=mock_client)

    mock_mlflow_run = mocker.Mock()
    converter_mock = mocker.patch(
        "mlflow_faculty.tracking.faculty_run_to_mlflow_run",
        return_value=mock_mlflow_run,
    )

    store = FacultyRestStore(STORE_URI)
    run = store.get_run(RUN_UUID_HEX_STR)

    assert run == mock_mlflow_run

    mock_client.get_run.assert_called_once_with(PROJECT_ID, RUN_UUID)
    converter_mock.assert_called_once_with(FACULTY_RUN)


def test_get_run_client_error(mocker):
    mock_client = mocker.Mock()
    mock_client.get_run.side_effect = HttpError(
        mocker.Mock(), "Experiment run with ID _ not found in project _"
    )
    mocker.patch("faculty.client", return_value=mock_client)

    store = FacultyRestStore(STORE_URI)

    with pytest.raises(
        MlflowException,
        match="Experiment run with ID _ not found in project _",
    ):
        store.get_run(RUN_UUID_HEX_STR)


def test_get_run_invalid_run_id(mocker):
    mock_client = mocker.Mock()
    mocker.patch("faculty.client", return_value=mock_client)

    store = FacultyRestStore(STORE_URI)

    with pytest.raises(ValueError):
        store.get_run("invalid-run-id")


@pytest.mark.parametrize(
    "run_status, faculty_run_status",
    [
        ("RUNNING", FacultyExperimentRunStatus.RUNNING),
        ("FINISHED", FacultyExperimentRunStatus.FINISHED),
        ("FAILED", FacultyExperimentRunStatus.FAILED),
        ("SCHEDULED", FacultyExperimentRunStatus.SCHEDULED),
        ("KILLED", FacultyExperimentRunStatus.KILLED),
        (RunStatus.RUNNING, FacultyExperimentRunStatus.RUNNING),
        (RunStatus.FINISHED, FacultyExperimentRunStatus.FINISHED),
        (RunStatus.FAILED, FacultyExperimentRunStatus.FAILED),
        (RunStatus.SCHEDULED, FacultyExperimentRunStatus.SCHEDULED),
        (RunStatus.KILLED, FacultyExperimentRunStatus.KILLED),
    ],
)
def test_update_run_info(mocker, run_status, faculty_run_status):
    faculty_run = mocker.Mock()
    mock_client = mocker.Mock()
    mock_client.update_run_info.return_value = faculty_run
    mocker.patch("faculty.client", return_value=mock_client)

    mlflow_run_info = mocker.Mock()
    mlflow_run = mocker.Mock()
    mlflow_run.info = mlflow_run_info
    run_converter_mock = mocker.patch(
        "mlflow_faculty.tracking.faculty_run_to_mlflow_run",
        return_value=mlflow_run,
    )

    store = FacultyRestStore(STORE_URI)

    returned_run_info = store.update_run_info(
        RUN_UUID_HEX_STR, run_status, RUN_ENDED_AT_MILLISECONDS
    )

    mock_client.update_run_info.assert_called_once_with(
        PROJECT_ID, RUN_UUID, faculty_run_status, RUN_ENDED_AT
    )
    run_converter_mock.assert_called_once_with(faculty_run)
    assert returned_run_info == mlflow_run_info


def test_update_run_info_client_error(mocker):
    mock_client = mocker.Mock()
    mock_client.update_run_info.side_effect = HttpError(
        mocker.Mock(), "Experiment run with ID _ not found in project _"
    )
    mocker.patch("faculty.client", return_value=mock_client)

    store = FacultyRestStore(STORE_URI)

    with pytest.raises(
        MlflowException,
        match="Experiment run with ID _ not found in project _",
    ):
        store.update_run_info(
            RUN_UUID_HEX_STR, "RUNNING", RUN_ENDED_AT_MILLISECONDS
        )


def test_update_run_info_invalid_run_id(mocker):
    mock_client = mocker.Mock()
    mocker.patch("faculty.client", return_value=mock_client)

    store = FacultyRestStore(STORE_URI)

    with pytest.raises(ValueError):
        store.update_run_info(
            "invalid-run-id", RunStatus.RUNNING, RUN_ENDED_AT_MILLISECONDS
        )


def test_delete_run(mocker):
    mock_client = mocker.Mock()
    mock_client.delete_runs.return_value = DeleteExperimentRunsResponse(
        deleted_run_ids=[RUN_UUID], conflicted_run_ids=[]
    )
    mocker.patch("faculty.client", return_value=mock_client)

    store = FacultyRestStore(STORE_URI)
    store.delete_run(RUN_UUID_HEX_STR)

    mock_client.delete_runs.assert_called_once_with(PROJECT_ID, [RUN_UUID])


@pytest.mark.parametrize(
    "response, message",
    [
        (
            DeleteExperimentRunsResponse(
                deleted_run_ids=[], conflicted_run_ids=[]
            ),
            "Could not delete non-existent run",
        ),
        (
            DeleteExperimentRunsResponse(
                deleted_run_ids=[], conflicted_run_ids=[RUN_UUID]
            ),
            "Could not delete already-deleted run",
        ),
    ],
)
def test_delete_run_failures(mocker, response, message):
    mock_client = mocker.Mock()
    mock_client.delete_runs.return_value = response
    mocker.patch("faculty.client", return_value=mock_client)

    store = FacultyRestStore(STORE_URI)
    with pytest.raises(MlflowException, match=message):
        store.delete_run(RUN_UUID_HEX_STR)


def test_delete_run_client_error(mocker):
    mock_client = mocker.Mock()
    mock_client.delete_runs.side_effect = HttpError(mocker.Mock(), "An error")
    mocker.patch("faculty.client", return_value=mock_client)

    store = FacultyRestStore(STORE_URI)
    with pytest.raises(MlflowException, match="An error"):
        store.delete_run(RUN_UUID_HEX_STR)


def test_delete_run_invalid_run_id(mocker):
    mock_client = mocker.Mock()
    mocker.patch("faculty.client", return_value=mock_client)

    store = FacultyRestStore(STORE_URI)

    with pytest.raises(ValueError):
        store.delete_run("invalid-run-id")


def test_restore_run(mocker):
    mock_client = mocker.Mock()
    mock_client.restore_runs.return_value = RestoreExperimentRunsResponse(
        restored_run_ids=[RUN_UUID], conflicted_run_ids=[]
    )
    mocker.patch("faculty.client", return_value=mock_client)

    store = FacultyRestStore(STORE_URI)
    store.restore_run(RUN_UUID_HEX_STR)

    mock_client.restore_runs.assert_called_once_with(PROJECT_ID, [RUN_UUID])


@pytest.mark.parametrize(
    "response, message",
    [
        (
            RestoreExperimentRunsResponse(
                restored_run_ids=[], conflicted_run_ids=[]
            ),
            "Could not restore non-existent run",
        ),
        (
            RestoreExperimentRunsResponse(
                restored_run_ids=[], conflicted_run_ids=[RUN_UUID]
            ),
            "Could not restore already-active run",
        ),
    ],
)
def test_restore_run_failures(mocker, response, message):
    mock_client = mocker.Mock()
    mock_client.restore_runs.return_value = response
    mocker.patch("faculty.client", return_value=mock_client)

    store = FacultyRestStore(STORE_URI)
    with pytest.raises(MlflowException, match=message):
        store.restore_run(RUN_UUID_HEX_STR)


def test_restore_run_client_error(mocker):
    mock_client = mocker.Mock()
    mock_client.restore_runs.side_effect = HttpError(mocker.Mock(), "An error")
    mocker.patch("faculty.client", return_value=mock_client)

    store = FacultyRestStore(STORE_URI)
    with pytest.raises(MlflowException, match="An error"):
        store.restore_run(RUN_UUID_HEX_STR)


def test_restore_run_invalid_run_id(mocker):
    mock_client = mocker.Mock()
    mocker.patch("faculty.client", return_value=mock_client)

    store = FacultyRestStore(STORE_URI)

    with pytest.raises(ValueError):
        store.restore_run("invalid-run-id")


def test_get_metric_history(mocker):
    metric_key = "metric_key"
    first_faculty_metric = mocker.Mock()
    second_faculty_metric = mocker.Mock()

    mock_client = mocker.Mock(
        get_metric_history=mocker.Mock(
            return_value=[first_faculty_metric, second_faculty_metric]
        )
    )
    mocker.patch("faculty.client", return_value=mock_client)

    first_mlflow_metric = mocker.Mock()
    second_mlflow_metric = mocker.Mock()
    metric_converter = mocker.patch(
        "mlflow_faculty.tracking.faculty_metric_to_mlflow_metric",
        side_effect=[first_mlflow_metric, second_mlflow_metric],
    )

    store = FacultyRestStore(STORE_URI)

    returned_metric_history = store.get_metric_history(
        RUN_UUID_HEX_STR, metric_key
    )

    mock_client.get_metric_history.assert_called_once_with(
        PROJECT_ID, RUN_UUID, metric_key
    )
    metric_converter.assert_has_calls(
        [mocker.call(first_faculty_metric), mocker.call(second_faculty_metric)]
    )
    assert returned_metric_history == [
        first_mlflow_metric,
        second_mlflow_metric,
    ]


def test_get_metric_history_client_error(mocker):
    mock_client = mocker.Mock()
    mock_client.get_metric_history.side_effect = HttpError(
        mocker.Mock(), "Dummy client error."
    )
    mocker.patch("faculty.client", return_value=mock_client)

    store = FacultyRestStore(STORE_URI)

    with pytest.raises(MlflowException, match="Dummy client error."):
        store.get_metric_history(RUN_UUID_HEX_STR, "metric-key")


def test_get_metric_history_invalid_run_id(mocker):
    mock_client = mocker.Mock()
    mocker.patch("faculty.client", return_value=mock_client)

    store = FacultyRestStore(STORE_URI)

    with pytest.raises(ValueError):
        store.get_metric_history("invalid-run-id", "metric-key")


def test_search_runs(mocker):
    mock_faculty_runs = [mocker.Mock(), mocker.Mock(), mocker.Mock()]
    list_page_1 = ListExperimentRunsResponse(
        runs=[mock_faculty_runs[0], mock_faculty_runs[1]],
        pagination=Pagination(
            start=0, size=2, previous=None, next=Page(start=2, limit=1)
        ),
    )
    list_page_2 = ListExperimentRunsResponse(
        runs=[mock_faculty_runs[2]],
        pagination=Pagination(
            start=2, size=1, previous=Page(start=0, limit=2), next=None
        ),
    )

    mock_client = mocker.Mock()
    mock_client.list_runs.side_effect = [list_page_1, list_page_2]
    mocker.patch("faculty.client", return_value=mock_client)

    mock_mlflow_runs = [mocker.Mock(), mocker.Mock(), mocker.Mock()]
    converter_mock = mocker.patch(
        "mlflow_faculty.tracking.faculty_run_to_mlflow_run",
        side_effect=mock_mlflow_runs,
    )

    lifecycle_stage_1 = mocker.Mock()
    lifecycle_stage_2 = mocker.Mock()
    mock_faculty_lifecycle_stages = [lifecycle_stage_1, lifecycle_stage_2]
    mocker.patch(
        "mlflow_faculty.tracking.mlflow_viewtype_to_faculty_lifecycle_stage",
        side_effect=mock_faculty_lifecycle_stages,
    )

    store = FacultyRestStore(STORE_URI)
    runs = store.search_runs(experiment_ids=None)

    assert runs == mock_mlflow_runs

    mock_client.list_runs.assert_has_calls(
        [
            mocker.call(
                PROJECT_ID,
                experiment_ids=None,
                lifecycle_stage=lifecycle_stage_1,
            ),
            mocker.call(
                PROJECT_ID,
                experiment_ids=None,
                lifecycle_stage=lifecycle_stage_2,
                start=2,
                limit=1,
            ),
        ]
    )
    converter_mock.assert_has_calls(
        [
            mocker.call(mock_faculty_runs[0]),
            mocker.call(mock_faculty_runs[1]),
            mocker.call(mock_faculty_runs[2]),
        ]
    )


def test_search_runs_empty_page(mocker):
    list_page = ListExperimentRunsResponse(
        runs=[],
        pagination=Pagination(start=0, size=0, previous=None, next=None),
    )

    mock_client = mocker.Mock()
    mock_client.list_runs.side_effect = [list_page]
    mocker.patch("faculty.client", return_value=mock_client)

    mock_faculty_lifecycle_stage = mocker.Mock()
    mocker.patch(
        "mlflow_faculty.tracking.mlflow_viewtype_to_faculty_lifecycle_stage",
        return_value=mock_faculty_lifecycle_stage,
    )

    store = FacultyRestStore(STORE_URI)
    runs = store.search_runs(experiment_ids=None)

    assert runs == []
    mock_client.list_runs.assert_called_once_with(
        PROJECT_ID,
        experiment_ids=None,
        lifecycle_stage=mock_faculty_lifecycle_stage,
    )


def test_search_runs_filter_by_experiment(mocker):
    list_page = ListExperimentRunsResponse(
        runs=[],
        pagination=Pagination(start=0, size=0, previous=None, next=None),
    )

    mock_client = mocker.Mock()
    mock_client.list_runs.side_effect = [list_page]
    mocker.patch("faculty.client", return_value=mock_client)

    mock_faculty_lifecycle_stage = mocker.Mock()
    mocker.patch(
        "mlflow_faculty.tracking.mlflow_viewtype_to_faculty_lifecycle_stage",
        return_value=mock_faculty_lifecycle_stage,
    )

    store = FacultyRestStore(STORE_URI)
    runs = store.search_runs(experiment_ids=[123, 456])

    assert runs == []
    mock_client.list_runs.assert_called_once_with(
        PROJECT_ID,
        experiment_ids=[123, 456],
        lifecycle_stage=mock_faculty_lifecycle_stage,
    )


def test_search_runs_with_max_runs(mocker):
    mock_faculty_runs = [
        mocker.Mock(),
        mocker.Mock(),
        mocker.Mock(),
        mocker.Mock(),
    ]
    list_page_1 = ListExperimentRunsResponse(
        runs=[mock_faculty_runs[0], mock_faculty_runs[1]],
        pagination=Pagination(
            start=0, size=2, previous=None, next=Page(start=2, limit=2)
        ),
    )
    list_page_2 = ListExperimentRunsResponse(
        runs=[mock_faculty_runs[2], mock_faculty_runs[3]],
        pagination=Pagination(
            start=2, size=2, previous=Page(start=0, limit=2), next=None
        ),
    )

    mock_client = mocker.Mock()
    mock_client.list_runs.side_effect = [list_page_1, list_page_2]
    mocker.patch("faculty.client", return_value=mock_client)

    mock_mlflow_runs = [
        mocker.Mock(),
        mocker.Mock(),
        mocker.Mock(),
        mocker.Mock(),
    ]
    converter_mock = mocker.patch(
        "mlflow_faculty.tracking.faculty_run_to_mlflow_run",
        side_effect=mock_mlflow_runs,
    )

    lifecycle_stage_1 = mocker.Mock()
    lifecycle_stage_2 = mocker.Mock()
    mock_faculty_lifecycle_stages = [lifecycle_stage_1, lifecycle_stage_2]
    mocker.patch(
        "mlflow_faculty.tracking.mlflow_viewtype_to_faculty_lifecycle_stage",
        side_effect=mock_faculty_lifecycle_stages,
    )

    store = FacultyRestStore(STORE_URI)
    runs = store.search_runs(experiment_ids=None, max_results=3)

    assert runs == mock_mlflow_runs[:3]

    mock_client.list_runs.assert_has_calls(
        [
            mocker.call(
                PROJECT_ID,
                experiment_ids=None,
                lifecycle_stage=lifecycle_stage_1,
            ),
            mocker.call(
                PROJECT_ID,
                experiment_ids=None,
                lifecycle_stage=lifecycle_stage_2,
                start=2,
                limit=2,
            ),
        ]
    )

    converter_mock.assert_has_calls(
        [
            mocker.call(mock_faculty_runs[0]),
            mocker.call(mock_faculty_runs[1]),
            mocker.call(mock_faculty_runs[2]),
        ]
    )
    assert (
        mocker.call(mock_faculty_runs[3]) not in converter_mock.call_args_list
    )


def test_search_runs_client_error(mocker):
    mock_client = mocker.Mock()
    mock_client.list_runs.side_effect = HttpError(
        mocker.Mock(), "Dummy client error."
    )
    mocker.patch("faculty.client", return_value=mock_client)
    mocker.patch(
        "mlflow_faculty.tracking.mlflow_viewtype_to_faculty_lifecycle_stage",
        return_value=mocker.Mock(),
    )

    store = FacultyRestStore(STORE_URI)

    with pytest.raises(MlflowException, match="Dummy client error."):
        store.search_runs([123])


def test_search_runs_invalid_experiment_id(mocker):
    mock_client = mocker.Mock()
    mocker.patch("faculty.client", return_value=mock_client)

    store = FacultyRestStore(STORE_URI)

    with pytest.raises(ValueError):
        store.search_runs(
            ["invalid-experiment-id", "invalid-experiment-id"], None, None
        )


def test_log_batch(mocker):
    mock_client = mocker.Mock()
    mocker.patch("faculty.client", return_value=mock_client)

    mlflow_metric = mocker.Mock()
    metric_converter_mock = mocker.patch(
        "mlflow_faculty.tracking.mlflow_metric_to_faculty_metric",
        return_value=mlflow_metric,
    )
    mlflow_param = mocker.Mock()
    param_converter_mock = mocker.patch(
        "mlflow_faculty.tracking.mlflow_param_to_faculty_param",
        return_value=mlflow_param,
    )
    mlflow_tag = mocker.Mock()
    tag_converter_mock = mocker.patch(
        "mlflow_faculty.tracking.mlflow_tag_to_faculty_tag",
        return_value=mlflow_tag,
    )

    store = FacultyRestStore(STORE_URI)
    store.log_batch(
        run_id=RUN_UUID_HEX_STR,
        metrics=[MLFLOW_METRIC],
        params=[MLFLOW_PARAM],
        tags=[MLFLOW_TAG],
    )

    mock_client.log_run_data.assert_called_once_with(
        PROJECT_ID,
        RUN_UUID,
        metrics=[mlflow_metric],
        params=[mlflow_param],
        tags=[mlflow_tag],
    )
    metric_converter_mock.assert_called_once_with(MLFLOW_METRIC)
    param_converter_mock.assert_called_once_with(MLFLOW_PARAM)
    tag_converter_mock.assert_called_once_with(MLFLOW_TAG)


def test_log_batch_empty(mocker):
    mock_client = mocker.Mock()
    mocker.patch("faculty.client", return_value=mock_client)

    store = FacultyRestStore(STORE_URI)
    store.log_batch(RUN_UUID_HEX_STR)

    mock_client.log_run_data.assert_called_once_with(
        PROJECT_ID, RUN_UUID, metrics=[], params=[], tags=[]
    )


def test_log_batch_param_conflict(mocker):
    mock_client = mocker.Mock()
    exception = faculty.clients.experiment.ParamConflict(
        message="message", conflicting_params=["param-key"]
    )

    mock_client.log_run_data.side_effect = exception
    mocker.patch("faculty.client", return_value=mock_client)

    store = FacultyRestStore(STORE_URI)

    with pytest.raises(MlflowException, match="param-key"):
        store.log_batch(RUN_UUID_HEX_STR)
    mock_client.log_run_data.assert_called_once_with(
        PROJECT_ID, RUN_UUID, metrics=[], params=[], tags=[]
    )


def test_log_batch_error(mocker):
    mock_client = mocker.Mock()
    exception = HttpError(
        mocker.Mock(), error="error_message", error_code="some_error_code"
    )

    mock_client.log_run_data.side_effect = exception
    mocker.patch("faculty.client", return_value=mock_client)

    store = FacultyRestStore(STORE_URI)

    with pytest.raises(MlflowException, match="error_message"):
        store.log_batch(RUN_UUID_HEX_STR)
    mock_client.log_run_data.assert_called_once_with(
        PROJECT_ID, RUN_UUID, metrics=[], params=[], tags=[]
    )


def test_log_batch_invalid_run_id(mocker):
    mock_client = mocker.Mock()
    mocker.patch("faculty.client", return_value=mock_client)

    store = FacultyRestStore(STORE_URI)

    with pytest.raises(ValueError):
        store.log_batch("invalid-run-id")


def test_log_metric(mocker):
    mock_client = mocker.Mock()
    mocker.patch("faculty.client", return_value=mock_client)

    mlflow_metric = mocker.Mock()
    metric_converter_mock = mocker.patch(
        "mlflow_faculty.tracking.mlflow_metric_to_faculty_metric",
        return_value=mlflow_metric,
    )

    store = FacultyRestStore(STORE_URI)
    store.log_metric(RUN_UUID_HEX_STR, MLFLOW_METRIC)

    mock_client.log_run_data.assert_called_once_with(
        PROJECT_ID, RUN_UUID, metrics=[mlflow_metric], params=[], tags=[]
    )
    metric_converter_mock.assert_called_once_with(MLFLOW_METRIC)


def test_log_param(mocker):
    mock_client = mocker.Mock()
    mocker.patch("faculty.client", return_value=mock_client)

    mlflow_param = mocker.Mock()
    param_converter_mock = mocker.patch(
        "mlflow_faculty.tracking.mlflow_param_to_faculty_param",
        return_value=mlflow_param,
    )

    store = FacultyRestStore(STORE_URI)
    store.log_param(RUN_UUID_HEX_STR, MLFLOW_PARAM)

    mock_client.log_run_data.assert_called_once_with(
        PROJECT_ID, RUN_UUID, metrics=[], params=[mlflow_param], tags=[]
    )
    param_converter_mock.assert_called_once_with(MLFLOW_PARAM)


def test_set_tag(mocker):
    mock_client = mocker.Mock()
    mocker.patch("faculty.client", return_value=mock_client)

    mlflow_tag = mocker.Mock()
    tag_converter_mock = mocker.patch(
        "mlflow_faculty.tracking.mlflow_tag_to_faculty_tag",
        return_value=mlflow_tag,
    )

    store = FacultyRestStore(STORE_URI)
    store.set_tag(RUN_UUID_HEX_STR, MLFLOW_TAG)

    mock_client.log_run_data.assert_called_once_with(
        PROJECT_ID, RUN_UUID, metrics=[], params=[], tags=[mlflow_tag]
    )
    tag_converter_mock.assert_called_once_with(MLFLOW_TAG)


def test_delete_experiment(mocker):
    mock_client = mocker.Mock()
    mocker.patch("faculty.client", return_value=mock_client)

    store = FacultyRestStore(STORE_URI)
    store.delete_experiment(EXPERIMENT_ID)

    mock_client.delete.assert_called_once_with(PROJECT_ID, EXPERIMENT_ID)


def test_delete_experiment_client_error(mocker):
    mock_client = mocker.Mock()
    mock_client.delete.side_effect = HttpError(
        mocker.Mock(), "Experiment with ID _ not found in project _"
    )
    mocker.patch("faculty.client", return_value=mock_client)

    store = FacultyRestStore(STORE_URI)

    with pytest.raises(
        MlflowException, match="Experiment with ID _ not found in project _"
    ):
        store.delete_experiment(EXPERIMENT_ID)

    mock_client.delete.assert_called_once_with(PROJECT_ID, EXPERIMENT_ID)


def test_delete_experiment_invalid_experiment_id(mocker):
    mock_client = mocker.Mock()
    mocker.patch("faculty.client", return_value=mock_client)

    store = FacultyRestStore(STORE_URI)

    with pytest.raises(ValueError):
        store.delete_experiment("invalid-experiment-id")


def test_restore_experiment(mocker):
    mock_client = mocker.Mock()
    mocker.patch("faculty.client", return_value=mock_client)

    store = FacultyRestStore(STORE_URI)
    store.restore_experiment(EXPERIMENT_ID)

    mock_client.restore.assert_called_once_with(PROJECT_ID, EXPERIMENT_ID)


def test_restore_experiment_client_error(mocker):
    mock_client = mocker.Mock()
    mock_client.restore.side_effect = HttpError(
        mocker.Mock(), "Experiment with ID _ not found in project _"
    )
    mocker.patch("faculty.client", return_value=mock_client)

    store = FacultyRestStore(STORE_URI)

    with pytest.raises(
        MlflowException, match="Experiment with ID _ not found in project _"
    ):
        store.restore_experiment(EXPERIMENT_ID)

    mock_client.restore.assert_called_once_with(PROJECT_ID, EXPERIMENT_ID)


def test_restore_experiment_invalid_experiment_id(mocker):
    mock_client = mocker.Mock()
    mocker.patch("faculty.client", return_value=mock_client)

    store = FacultyRestStore(STORE_URI)

    with pytest.raises(ValueError):
        store.restore_experiment("invalid-experiment-id")



# Copyright 2019 Faculty Science Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from uuid import uuid4

import pytest

from mlflow_faculty.artifacts import (
    _S3ArtifactRepositoryWithClientOverride,
    FacultyDatasetsArtifactRepository,
)


PROJECT_ID = uuid4()
ARTIFACT_URI = "faculty-datasets:{}/path/in/datasets".format(PROJECT_ID)
S3_BUCKET_NAME = "faculty-datasets-s3-bucket"
S3_URI = "s3://{}/{}/path/in/datasets".format(S3_BUCKET_NAME, PROJECT_ID)


@pytest.fixture
def mock_session(mocker):
    session = mocker.Mock()
    session.bucket.return_value = S3_BUCKET_NAME
    mocker.patch("faculty.datasets.session.get", return_value=session)
    return session


@pytest.fixture
def mock_s3_repo(mocker):
    s3_repo = mocker.Mock()
    mocker.patch(
        "mlflow_faculty.artifacts._S3ArtifactRepositoryWithClientOverride",
        return_value=s3_repo,
    )
    return s3_repo


def test_s3_repo_with_client_override(mocker):

    artifact_uri = mocker.Mock()
    s3_client_factory = mocker.Mock()

    repo = _S3ArtifactRepositoryWithClientOverride(
        artifact_uri, s3_client_factory
    )

    assert repo.artifact_uri == artifact_uri
    assert repo._get_s3_client() == s3_client_factory.return_value


@pytest.mark.parametrize("suffix", ["", "/"])
def test_faculty_repo(mocker, mock_session, suffix):

    s3_repo = mocker.Mock()
    s3_repo_init = mocker.patch(
        "mlflow_faculty.artifacts._S3ArtifactRepositoryWithClientOverride",
        return_value=s3_repo,
    )

    repo = FacultyDatasetsArtifactRepository(ARTIFACT_URI + suffix)

    assert repo._s3_repository == s3_repo

    # Check arguments that the S3 repo was built with
    passed_uri, passed_client_factory = s3_repo_init.call_args[0]
    # The first argument should be the correct S3 URI
    assert passed_uri == S3_URI
    # The second argument should be a function that builds an S3 client with
    # the correct project ID
    assert passed_client_factory() == mock_session.s3_client.return_value
    mock_session.s3_client.assert_called_once_with(PROJECT_ID)


@pytest.mark.parametrize(
    "uri, message",
    [
        ("no/schema", "Not a Faculty datasets URI"),
        ("wrong-schema:", "Not a Faculty datasets URI"),
        (
            "faculty-datasets://{}/path/in/datasets".format(PROJECT_ID),
            "Invalid URI.*Did you mean '{}'".format(ARTIFACT_URI),
        ),
        ("faculty-datasets:invalid-uri", "is not a valid UUID"),
    ],
    ids=["No schema", "Wrong schema", "Double slash", "Invalid UUID"],
)
def test_faculty_repo_invalid_uri(uri, message):
    with pytest.raises(ValueError, match=message):
        FacultyDatasetsArtifactRepository(uri)


def test_faculty_repo_get_path_module(mock_session, mock_s3_repo):
    repo = FacultyDatasetsArtifactRepository(ARTIFACT_URI)
    assert repo.get_path_module() == mock_s3_repo.get_path_module.return_value
    mock_s3_repo.get_path_module.assert_called_once_with()


def test_faculty_repo_log_artifact(mocker, mock_session, mock_s3_repo):
    local_file = mocker.Mock()
    artifact_path = mocker.Mock()
    repo = FacultyDatasetsArtifactRepository(ARTIFACT_URI)

    returned = repo.log_artifact(local_file, artifact_path)

    assert returned == mock_s3_repo.log_artifact.return_value
    mock_s3_repo.log_artifact.assert_called_once_with(
        local_file, artifact_path
    )


def test_faculty_repo_log_artifacts(mocker, mock_session, mock_s3_repo):
    local_file = mocker.Mock()
    artifact_path = mocker.Mock()
    repo = FacultyDatasetsArtifactRepository(ARTIFACT_URI)

    returned = repo.log_artifacts(local_file, artifact_path)

    assert returned == mock_s3_repo.log_artifacts.return_value
    mock_s3_repo.log_artifacts.assert_called_once_with(
        local_file, artifact_path
    )


def test_faculty_repo_list_artifacts(mocker, mock_session, mock_s3_repo):
    path = mocker.Mock()
    repo = FacultyDatasetsArtifactRepository(ARTIFACT_URI)

    returned = repo.list_artifacts(path)

    assert returned == mock_s3_repo.list_artifacts.return_value
    mock_s3_repo.list_artifacts.assert_called_once_with(path)


def test_faculty_repo_download_file(mocker, mock_session, mock_s3_repo):
    remote_file_path = mocker.Mock()
    local_path = mocker.Mock()
    repo = FacultyDatasetsArtifactRepository(ARTIFACT_URI)

    returned = repo._download_file(remote_file_path, local_path)

    assert returned == mock_s3_repo._download_file.return_value
    mock_s3_repo._download_file.assert_called_once_with(
        remote_file_path, local_path
    )
